package com.lts.objectdetector;

import com.facebook.react.uimanager.NativeViewHierarchyManager;
import com.facebook.react.uimanager.UIManagerModule;
import com.facebook.react.uimanager.UIBlock;
import com.facebook.react.bridge.ReadableMap;
import com.facebook.react.bridge.Promise;
import com.facebook.react.bridge.ReactApplicationContext;
import com.facebook.react.bridge.ReactContextBaseJavaModule;
import com.facebook.react.bridge.ReactMethod;

public class ObjectDetectorModule extends ReactContextBaseJavaModule {

    private static final String TAG = ObjectDetectorModule.class.getSimpleName();

    public ObjectDetectorModule(ReactApplicationContext reactContext) {
        super(reactContext);
    }

    @Override
    public String getName() {
        return "ObjectDetectorModule";
    }

    private ReadableMap mVideoOptions;
	
    @ReactMethod
    public void takePicture(final ReadableMap options, final int viewTag, final Promise promise) {
      final ReactApplicationContext context = getReactApplicationContext();
      UIManagerModule uiManager = context.getNativeModule(UIManagerModule.class);
      uiManager.addUIBlock(new UIBlock() {
          @Override
          public void execute(NativeViewHierarchyManager nativeViewHierarchyManager) {
              ObjectDetectorView cameraView = (ObjectDetectorView) nativeViewHierarchyManager.resolveView(viewTag);
              try {				  
			      TakePicBlock takePicBlock = new TakePicBlock(options, promise);
                  cameraView.takePicture(takePicBlock);
              } 
		      catch (Exception e) {
                  promise.reject("E_CAMERA_BAD_VIEWTAG", "takePicture: Expected a Camera component");
              }
          }
      });
    }

    @ReactMethod
    public void initCamera(final int viewTag, final Promise promise) {
        final ReactApplicationContext context = getReactApplicationContext();
        UIManagerModule uiManager = context.getNativeModule(UIManagerModule.class);
        uiManager.addUIBlock(new UIBlock() {
            @Override
            public void execute(NativeViewHierarchyManager nativeViewHierarchyManager) {
                ObjectDetectorView cameraView = (ObjectDetectorView) nativeViewHierarchyManager.resolveView(viewTag);
                try {
                    InitCameraBlock initCameraBlock = new InitCameraBlock(cameraView, promise);
                    cameraView.initCamera(initCameraBlock);
                }
                catch (Exception e) {
                    promise.reject("E_CAMERA_BAD_VIEWTAG", "initCamera: Expected a Camera component");
                }
            }
        });
    }
}
