package com.lts.objectdetector;

import com.facebook.react.bridge.Promise;
import com.facebook.react.uimanager.ThemedReactContext;
import com.facebook.react.bridge.WritableNativeMap;
import com.facebook.react.bridge.WritableMap;
import com.facebook.react.bridge.WritableArray;
import com.facebook.react.bridge.ReadableMap;
import com.facebook.react.modules.core.DeviceEventManagerModule;

import android.Manifest;
import android.content.pm.PackageManager;
import android.content.res.Resources;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.net.Uri;
import android.os.AsyncTask;
import android.os.Environment;
import android.util.Log;
import android.view.SurfaceHolder;
import android.view.TextureView;
import android.util.Base64;
import android.view.OrientationEventListener;

import androidx.core.content.ContextCompat;

import org.opencv.calib3d.Calib3d;
import org.opencv.core.DMatch;
import org.opencv.core.KeyPoint;
import org.opencv.core.MatOfDMatch;
import org.opencv.core.MatOfKeyPoint;
import org.opencv.core.MatOfPoint;
import org.opencv.features2d.BFMatcher;
import org.opencv.features2d.DescriptorMatcher;
import org.opencv.features2d.Features2d;
import org.opencv.features2d.ORB;
import org.opencv.android.Utils;
import org.opencv.core.Mat;
import org.opencv.core.MatOfPoint2f;
import org.opencv.core.Point;
import org.opencv.core.Scalar;
import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Size;
import org.opencv.imgcodecs.Imgcodecs;
import org.opencv.imgproc.Imgproc;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.Runnable;
import java.lang.ref.WeakReference;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;

class TakePicBlock implements Runnable {

	Mat mat;
	ReadableMap options;
	Promise promise;
	
   	public TakePicBlock(ReadableMap options, Promise promise) {
		this.options = options;
		this.promise = promise;
   	}

   	public void setMat(Mat mat) {
	   this.mat = mat;
   	}
   
   	public void run() {
        EngineUtils.getInstance().matToImage(mat, options.getString("filename"), promise);
   	}
}

class InitCameraBlock implements Runnable {

    Promise promise;
    ObjectDetectorView cameraView;

    public InitCameraBlock(ObjectDetectorView cameraView, Promise promise) {
        this.cameraView = cameraView;
        this.promise = promise;
    }

    public void run() {
        cameraView.setVisibility(TextureView.VISIBLE);

        if (this.promise != null) {
            WritableNativeMap result = new WritableNativeMap();
            result.putBoolean("cameraInitialized", true);
            promise.resolve(result);
        }
    }
}

public class ObjectDetectorView extends JavaCameraView implements CameraBridgeViewBase.CvCameraViewListener2 {

    private static final String TAG = ObjectDetectorView.class.getSimpleName();

    public  SurfaceHolder          mHolder;
    public  ThemedReactContext     mContext;

    // params
    private ReadableMap            mOverlay;
    private Mat                    mOverlayMat;
    private int                    mOverlayInterval = 0;
    private ReadableMap            mCvInvokeGroup;
    private int                    mCameraFacing;
    private boolean                mSuckUpFrames;

    private boolean                mUpdateOverlay      = false;
    private boolean                mSendFrameSize      = true;
    private int                    mRotation           = -1;
    private int                    mRecRotation        = -1;
    private long                   mCurrentMillis      = -1;
    private int                    mCurrOverlayIndex   = -1;
    private boolean				   mTakePicture	       = false;
	private TakePicBlock	       takePicBlock;
	private InitCameraBlock        initCameraBlock;
	
	// video and audio recording stuff
	private Mat                    mSampleImageMat     = null;
	private Scalar                 mOverlayColor;
	private float                  mOverlayTransparency = 0.2f;
	private int                    mNDetectionB4Take   = 5;

    private boolean                mDetecting          = false;
    private Mat                    mCurrentImage;
    private ORB                    mORB;
    private DescriptorMatcher      mDescriptorMatcher;
    private MatOfKeyPoint          msKP;
    private Mat                    msDes;
    private Mat                    mWarpMatrix         = null;
    private int                    mNDetected          = 0;
    private DetectObjectAsync      mAsyncTask          = null;
    private Mat                    mMask               = new Mat();
    private String                 mMainPackageName    = "";

    private int                    m_total = 0;

    public ObjectDetectorView(ThemedReactContext context, int cameraFacing) {
        super( context, cameraFacing);
        Log.d(TAG, "Creating and setting view");
        mCameraFacing = cameraFacing;
        mContext = context;

        this.setVisibility(TextureView.INVISIBLE);
        this.setCvCameraViewListener(this);

        System.loadLibrary("opencv_java3");

        mHolder = getHolder();
        mHolder.addCallback(this);

        // this is for older devices might as well keep it in here ...
        mHolder.setType(SurfaceHolder.SURFACE_TYPE_PUSH_BUFFERS);

        if (ContextCompat.checkSelfPermission(getContext(),
            	Manifest.permission.CAMERA) == PackageManager.PERMISSION_GRANTED) {
            this.initCameraBlock = new InitCameraBlock(this, null);
            this.initCameraBlock.run();
        }

        mORB = ORB.create(15000, 1.3f, 8, 10, 1, 2, ORB.HARRIS_SCORE, 31, 10);
        mDescriptorMatcher = BFMatcher.create(BFMatcher.BRUTEFORCE_HAMMING, true);
        mOverlayColor = new Scalar(0, 0, 255, 0.2);

        setMainPackageName(context.getApplicationContext().getPackageName());

        OrientationEventListener orientationEventListener = new OrientationEventListener(mContext) {
            @Override
            public void onOrientationChanged(int rotation) {
                if (((rotation >= 0) && (rotation <= 45)) || (rotation > 315)) {
                    mRotation = Core.ROTATE_90_CLOCKWISE;
                } else if ((rotation > 45) && (rotation <= 135)) {
                    mRotation = Core.ROTATE_180;
                } else if ((rotation > 135) && (rotation <= 225)) {
                    mRotation = Core.ROTATE_90_COUNTERCLOCKWISE;
                } else {
                    mRotation = -1;
                }
                Log.d(TAG, "orientation = " + mRotation);
            }
        };
        orientationEventListener.enable();
    }

    @Override
    public void surfaceCreated(SurfaceHolder holder) {
      super.surfaceCreated( holder);
      Log.d(TAG, "In surfaceCreated ...");
    }

    @Override
    public void surfaceChanged(SurfaceHolder holder, int format, int w, int h) {
        super.surfaceChanged( holder, format, w, h);

        Log.d(TAG, "In surfaceChanged ...");

        if (mHolder.getSurface() == null){
            Log.e(TAG, "In surfaceChanged surface is null ...");
            // preview surface does not exist
            return;
        }

        try {
            this.enableView();
        }
        catch (Exception e){
            Log.e("CameraPreview", "Error enabling camera preview: " + e.getMessage());
        }
    }

    // TODO: not sure if this is the right place for this ...
    public void changeFacing(int facing) {
        if (mCameraFacing != facing) {
            mCameraFacing = facing;
            // TODO: restart camera preview
            disableView();
            setCameraIndex(mCameraFacing);
            enableView();
        }
    }

    public void setCvInvokeGroup(ReadableMap cvinvoke) {
        //Log.d(TAG, "In setCvInvoke cvinvoke is: " + cvinvoke.toString());
        this.mCvInvokeGroup = cvinvoke;
    }

    public void setMainPackageName(String name) {
        mMainPackageName = name;
    }

    public void setSampleImage(String uri) {
        if (uri.contains("http")) {
            new AsyncTask<String, Void, String>(){
                @Override
                protected String doInBackground(String... params) {
                    try {
                        File dir = getContext().getCacheDir();

                        URL ulrn = new URL(params[0]);
                        HttpURLConnection con = (HttpURLConnection) ulrn.openConnection();
                        InputStream is = con.getInputStream();

                        String fileName = (params[1] == null) ? String.valueOf(System
                                .currentTimeMillis()) : params[1];

                        try {
                            OutputStream fOut = null;
                            File file = new File(dir.getAbsolutePath(), fileName);
                            if (file.exists()) {
                                file.delete();
                            }

                            fOut = new FileOutputStream(file);

                            int read = 0;
                            byte[] bytes = new byte[1024];

                            while ((read = is.read(bytes)) != -1) {
                                fOut.write(bytes, 0, read);
                            }
                            is.close();

                            fOut.flush();
                            fOut.close();

                            return file.getAbsolutePath();
                        } catch (Exception e) {
                            e.printStackTrace();
                        }

                    } catch (Exception e) {
                        e.printStackTrace();
                    }

                    return null;
                }

                @Override
                protected void onPostExecute(String path) {
                    super.onPostExecute(path);
                    mSampleImageMat = null;

                    File f = null;
                    if (path != null)
                        f = new File(path);

                    if (path != null && f.exists()) {
                        Bitmap bitmap = EngineUtils.loadBitmap(path);

                        if (bitmap != null) {
                            bitmap = EngineUtils.rotateBitmap(path, bitmap);
                        }

                        mSampleImageMat = new Mat();
                        Utils.bitmapToMat(bitmap, mSampleImageMat);
                        Imgproc.cvtColor(mSampleImageMat, mSampleImageMat, Imgproc.COLOR_RGB2GRAY);
                    }

                    getFeaturesOfSample();
                }
            }.execute(uri, "sampleImage.jpg");

        } else if (!uri.isEmpty()){
            // try to load from resource
            int resID = getResources().getIdentifier(uri , "drawable", mMainPackageName);
            String path = "android.resource://" + mMainPackageName + "/drawable/" + uri;
            Bitmap bitmap = BitmapFactory.decodeResource( getResources(), resID);

            if (bitmap != null) {
                Uri imageuri = Uri.parse(path);
                try {
                    InputStream stream = getContext().getContentResolver().openInputStream(imageuri);
                    bitmap = EngineUtils.rotateBitmap(stream, bitmap);
                } catch (Exception e) {
                    e.printStackTrace();
                }

                mSampleImageMat = new Mat();
                Utils.bitmapToMat(bitmap, mSampleImageMat);

                Imgproc.cvtColor(mSampleImageMat, mSampleImageMat, Imgproc.COLOR_BGR2GRAY);

                getFeaturesOfSample();
            } else {
                mSampleImageMat = null;
            }
        } else {
            mSampleImageMat = null;
        }
    }

    private void getFeaturesOfSample() {
        if (mSampleImageMat != null) {
            Imgproc.resize(mSampleImageMat, mSampleImageMat, new Size(250, 250));

            msKP = new MatOfKeyPoint();
            msDes = new Mat();

            mORB.setFirstLevel(0);
            mORB.detect(mSampleImageMat, msKP);
            mORB.compute(mSampleImageMat, msKP, msDes);
            mORB.setFirstLevel(1);

            ArrayList<Mat> descriptors = new ArrayList<>();
            descriptors.add(msDes);
            mDescriptorMatcher.add(descriptors);
            mDescriptorMatcher.train();
            Mat nImg = new Mat();
            Features2d.drawKeypoints(mSampleImageMat, msKP, nImg);
            EngineUtils.saveImageMat(nImg, "sample_kp.jpg");
        }
    }

    public void setOverlayColor(Scalar color) {
        mOverlayColor = color;
        if (color.val.length == 4) {
            mOverlayTransparency = (float)color.val[3] / 255.0f;
        } else {
            mOverlayTransparency = 1.0f;
        }
    }

    public void setDetectionCountBeforeCapture(int count) {
        if (count > 0)
            mNDetectionB4Take = count;
    }

    public void setOverlay(ReadableMap overlay) {
        //Log.d(TAG, "About to set overlay.");
        if (overlay != null) {
            mOverlay = overlay;
            int matIndex = overlay.getInt("matIndex");
            Mat overlayMat = (Mat)MatManager.getInstance().matAtIndex(matIndex);
            if (mOverlayMat == null) {
                mOverlayMat = new Mat(overlayMat.rows(), overlayMat.cols(), CvType.CV_8UC4);
            }
            overlayMat.copyTo(mOverlayMat);
            Scalar clearColor = Scalar.all(0);
            overlayMat.setTo(clearColor);
            MatManager.getInstance().setMat(matIndex, overlayMat);
            mUpdateOverlay = true;
        }
    }

    public void setOverlayInterval(int overlayInterval) {
        this.mOverlayInterval = overlayInterval;
    }

    public void onCameraViewStarted(int width, int height) {
        Log.d(TAG, "In onCameraViewStarted ... width is: " + width + " height is: " + height);
    }

    public void onCameraViewStopped() {
        Log.d(TAG, "In onCameraViewStopped ...");
        this.disableView();
    }

    private void rotateImage(Mat image) {
        if (mRotation != -1) {
            Core.rotate(image, image, mRotation);
        }
    }

    private double calcDistance(double centerX, double centerY, double pointX, double pointY) {
        double distX = pointX - centerX;
        double distY = pointY - centerY;

        distX = (distX < 0.0) ? -distX : distX;
        distY = (distY < 0.0) ? -distY : distY;
        return (distX + distY);
    }

    public Mat onCameraFrame(CvCameraViewFrame inputFrame) {
        Mat in = inputFrame.rgba();

        Mat ingray = inputFrame.gray();

        if (mCurrentMillis == -1) {
            mCurrentMillis = System.currentTimeMillis();
        }

        WritableMap fsResponse = new WritableNativeMap();
        //Log.d(TAG, "payload is: " + faceInfo);
        String frameSizeStr = "{\"frameSize\":{\"frameWidth\":" + (double)in.size().width + ",\"frameHeight\":" + (double)in.size().height + "}}";
        fsResponse.putString("payload", frameSizeStr);
        mContext.getJSModule(DeviceEventManagerModule.RCTDeviceEventEmitter.class)
            .emit("onFrameSize", fsResponse);

        if (mCameraFacing == 1) {
            Core.flip(in, in, 1);
            Core.flip(ingray, ingray, 1);
        }

        boolean sendCallbackData = false;

        if (mCvInvokeGroup != null) {
            long currMillis = System.currentTimeMillis();
            long diff = (currMillis - mCurrentMillis);
            if (diff >= mOverlayInterval) {
                mCurrentMillis = currMillis;
		        CvInvoke invoker = new CvInvoke(in, ingray);
		        WritableArray responseArr = invoker.parseInvokeMap(mCvInvokeGroup);
		        String lastCall = invoker.callback;
				int dstMatIndex = invoker.dstMatIndex;
				// TODO: move this into RNOpencv3Util class ...
		        if (lastCall != null && !lastCall.equals("") && dstMatIndex >= 0 && dstMatIndex < 1000) {
		            WritableMap response = new WritableNativeMap();
		            response.putArray("payload", responseArr);
		            mContext.getJSModule(DeviceEventManagerModule.RCTDeviceEventEmitter.class)
		                .emit(lastCall, response);
		        }
		        else {
		            // not necessarily error condition unless dstMatIndex >= 1000
		            if (dstMatIndex == 1000) {
		                Log.e(TAG, "SecurityException thrown attempting to invoke method.  Check your method name and parameters and make sure they are correct.");
		            }
		            else if (dstMatIndex == 1001) {
		                Log.e(TAG, "IllegalAccessException thrown attempting to invoke method.  Check your method name and parameters and make sure they are correct.");
		            }
		            else if (dstMatIndex == 1002) {
		                Log.e(TAG, "InvocationTargetException thrown attempting to invoke method.  Check your method name and parameters and make sure they are correct.");
		            }
		        }
            }
        }
		
        if (mOverlayMat != null) {
            Core.addWeighted(in, 1.0, mOverlayMat, 1.0, 0.0, in);
        }
		
		if (mTakePicture) {
			mTakePicture = false;
			this.takePicBlock.setMat(in);
			this.takePicBlock.run();
		}

//        m_total++;

        if (mSuckUpFrames) {
            // AKA bowel movement!
            Bitmap bm = Bitmap.createBitmap(in.cols(), in.rows(), Bitmap.Config.ARGB_8888);
            Utils.matToBitmap(in, bm);
            //writeImage(bm);
            String encodedData = EngineUtils.toBase64(bm, 60);
            WritableMap response = new WritableNativeMap();
            response.putString("payload", encodedData);
            mContext.getJSModule(DeviceEventManagerModule.RCTDeviceEventEmitter.class)
                .emit("onCameraFrame", response);
        }

        if (!mDetecting && mSampleImageMat != null) {
            mDetecting = true;
            mCurrentImage = in;
            mAsyncTask = new DetectObjectAsync(this, mORB, mDescriptorMatcher, msKP, msDes, mMask);
            mAsyncTask.execute(ingray);
        }

        if (mNDetected > 0) {
            Mat transOverlay = in.clone();

            Mat objCorners = new Mat(4, 1, CvType.CV_32FC2), sceneCorners = new Mat();
            float[] objCornersData = new float[(int) (objCorners.total() * objCorners.channels())];
            objCorners.get(0, 0, objCornersData);
            objCornersData[0] = 0;
            objCornersData[1] = 0;
            objCornersData[2] = mSampleImageMat.cols();
            objCornersData[3] = 0;
            objCornersData[4] = mSampleImageMat.cols();
            objCornersData[5] = mSampleImageMat.rows();
            objCornersData[6] = 0;
            objCornersData[7] = mSampleImageMat.rows();
            objCorners.put(0, 0, objCornersData);

            if (mWarpMatrix == null) return in;
            Core.perspectiveTransform(objCorners, sceneCorners, mWarpMatrix);
            float[] sceneCornersData = new float[(int) (sceneCorners.total() * sceneCorners.channels())];
            sceneCorners.get(0, 0, sceneCornersData);

            Point[] scene_corners = new Point[4];
            scene_corners[0] = new Point(sceneCornersData[0], sceneCornersData[1]);
            scene_corners[1] = new Point(sceneCornersData[2], sceneCornersData[3]);
            scene_corners[2] = new Point(sceneCornersData[4], sceneCornersData[5]);
            scene_corners[3] = new Point(sceneCornersData[6], sceneCornersData[7]);

            ArrayList<MatOfPoint> pts = new ArrayList<>();
            MatOfPoint Pt = new MatOfPoint();
            Pt.fromArray(scene_corners);
            pts.add(Pt);

            Imgproc.fillPoly(transOverlay, pts, mOverlayColor);
            Core.addWeighted(in, (1 - mOverlayTransparency), transOverlay, mOverlayTransparency, 0, transOverlay);

            return transOverlay;
        }

        return in;
    }

    public void initCamera(InitCameraBlock initCameraBlock) {
        this.initCameraBlock = initCameraBlock;
        this.initCameraBlock.run();
    }

    public void takePicture(TakePicBlock takePicBlock) {
        this.takePicBlock = takePicBlock;
		mTakePicture = true;
    }
	
    @Override
    protected boolean connectCamera(int width, int height) {
      boolean supVal = super.connectCamera( width, height);
      Log.d(TAG, "In connectCamera ...");
      return supVal;
    }

    /**
     * Disconnects and release the particular camera object being connected to this surface view.
     * Called when syncObject lock is held
     */
    @Override
    protected void disconnectCamera() {
      super.disconnectCamera( );
      mAsyncTask.cancel(true);
	  mNDetected = 0;
	  mDetecting = false;
      Log.d(TAG, "In disconnectCamera ...");
    }

    @Override
    protected boolean initializeCamera(int width, int height) {
        boolean result = super.initializeCamera(width, height);

        mMask = new Mat(mFrameHeight, mFrameWidth, CvType.CV_8U, Scalar.all(0));
        if (mFrameWidth < mFrameHeight) {
            Imgproc.rectangle(mMask, new Point(mFrameWidth / 6.0, mFrameHeight / 6.0),
                    new Point(mFrameWidth * 5 / 6.0, mFrameHeight * 5 / 6.0), new Scalar(255), -1);
        } else {
            Imgproc.rectangle(mMask, new Point(mFrameWidth / 6.0, mFrameHeight / 6.0),
                    new Point(mFrameWidth * 5 / 6.0, mFrameHeight * 5 / 6.0), new Scalar(255), -1);
        }
        return result;
    }

    private static class DetectObjectAsync extends AsyncTask<Mat, Void, Mat> {
        private WeakReference<ObjectDetectorView> viewReference;
        private Mat a_mask, sdes;
        private MatOfKeyPoint skp;
        private ORB a_orb;
        private DescriptorMatcher a_matcher;

        DetectObjectAsync(ObjectDetectorView view, ORB orb, DescriptorMatcher matcher, MatOfKeyPoint kp, Mat des, Mat mask) {
            viewReference = new WeakReference<>(view);
            a_orb = orb;
            a_matcher = matcher;
            a_mask = mask;
            skp = kp;
            sdes = des;
        }

        @Override
        protected Mat doInBackground(Mat... imgs) {
            Mat img = imgs[0];
            Imgproc.GaussianBlur(img, img, new Size(5, 5), 0);

            MatOfKeyPoint kp = new MatOfKeyPoint();
            Mat des = new Mat();

            a_orb.detectAndCompute(img, a_mask, kp, des);
//            ObjectDetectorView oview = viewReference.get();

//            Mat temp = new Mat();
//            Features2d.drawKeypoints(img, kp, temp);
//            EngineUtils.saveImageMat(temp, "tempimage_kp.jpg");

            MatOfDMatch matches = new MatOfDMatch();
            a_matcher.match(des, matches);
            DMatch[] matchesArray = matches.toArray();

//            Mat atemp = new Mat();
//            Features2d.drawMatches(oview.mSampleImageMat, skp, img, kp, matches, atemp);
//            EngineUtils.saveImageMat(atemp, "match.jpg");
            if (matchesArray.length < 4) {
                // not found
                return null;
            }

            Arrays.sort(matchesArray, new Comparator<DMatch>() {
                @Override
                public int compare(DMatch o1, DMatch o2) {
                    return Double.compare(o1.distance, o2.distance);
                }
            });

            LinkedList<Point> sPList = new LinkedList<>();
            LinkedList<Point> pList = new LinkedList<>();

            List<KeyPoint> sKPList = skp.toList();
            List<KeyPoint> kpList = kp.toList();

            for (int i=0; i<matchesArray.length && i<50; i++) {
                sPList.addLast(sKPList.get(matchesArray[i].trainIdx).pt);
                pList.addLast(kpList.get(matchesArray[i].queryIdx).pt);
            }

            MatOfPoint2f sPts = new MatOfPoint2f();
            MatOfPoint2f pts = new MatOfPoint2f();

            sPts.fromList(sPList);
            pts.fromList(pList);

            Mat H = Calib3d.findHomography(sPts, pts, Calib3d.RANSAC, 5.0);

            if (!H.empty() && EngineUtils.isNiceHomography(H)) {
                return H;
            }

            return null;
        }

        @Override
        protected void onPostExecute(Mat H) {
            super.onPostExecute(H);

            ObjectDetectorView view = viewReference.get();
            if (view == null) return;

            if (H != null) {
                view.mNDetected++;

                if (view.mNDetected >= view.mNDetectionB4Take) {
                    Mat invH = H.inv();
                    Mat dest = new Mat();
                    Imgproc.warpPerspective(view.mCurrentImage, dest, invH, view.mSampleImageMat.size());
                    Imgproc.resize(dest, dest, new Size(500, 500));

//                    Imgproc.cvtColor(aMat, aMat, Imgproc.COLOR_BGR2RGB);
                    Bitmap bm = Bitmap.createBitmap(dest.cols(), dest.rows(), Bitmap.Config.ARGB_8888);
                    Utils.matToBitmap(dest, bm);
                    String encodedData = EngineUtils.toBase64(bm, 80);
                    WritableMap response = new WritableNativeMap();
                    response.putString("payload", encodedData);
                    view.mContext.getJSModule(DeviceEventManagerModule.RCTDeviceEventEmitter.class)
                            .emit("onObjectCaptured", response);

                    view.mNDetected = 0;
                }

                view.mWarpMatrix = H.clone();
            } else {
                view.mWarpMatrix = null;
                view.mNDetected = 0;
            }

            view.mDetecting = false;
        }
    }
}
