package com.lts.objectdetector;

import com.facebook.react.uimanager.ThemedReactContext;
//import com.facebook.react.uimanager.ViewGroupManager;
import com.facebook.react.uimanager.SimpleViewManager;
import com.facebook.react.uimanager.annotations.ReactProp;
import com.facebook.react.bridge.ReadableMap;
import com.facebook.react.bridge.ReadableArray;
import com.facebook.react.common.MapBuilder;

import java.util.Map;
import javax.annotation.Nullable;

import android.util.Log;

public class ObjectDetectorViewManager extends SimpleViewManager<ObjectDetectorView> {

    private static final String TAG = ObjectDetectorViewManager.class.getSimpleName();

    public static final int CMD_OVERLAY_MAT = 1;

    @Override
    public String getName() {
        return "ObjectDetectorView";
    }

    @Override
    protected ObjectDetectorView createViewInstance(ThemedReactContext reactContext) {
        return new ObjectDetectorView(reactContext, -1);
    }

    @Override
    public Map<String,Integer> getCommandsMap() {
     Log.d(TAG, "View manager getCommandsMap:");
     return MapBuilder.of(
       "setOverlay",
       CMD_OVERLAY_MAT);
    }

    @Override
    public void receiveCommand(ObjectDetectorView view, int commandType, @Nullable ReadableArray args) {
        switch (commandType) {
            case CMD_OVERLAY_MAT: {
                view.setOverlay(args.getMap(0));
                return;
            }
            default:
                throw new IllegalArgumentException(String.format(
                    "Unsupported command %d received by %s.",
                    commandType,
                    getClass().getSimpleName()));

        }
    }

    @ReactProp(name = "facing")
    public void setFacing(ObjectDetectorView view, String facing) {
        view.changeFacing(facing.equals("front") ? 1 : -1);
    }

    @ReactProp(name = "sampleImageURI")
    public void setSampleImageURI(ObjectDetectorView view, String uri) {
        view.setSampleImage(uri);
    }

    @ReactProp(name = "cvinvoke")
    public void setCvInvoke(ObjectDetectorView view, ReadableMap cvinvoke) {
        view.setCvInvokeGroup(cvinvoke);
    }

    @ReactProp(name = "overlayInterval")
    public void setOverlayInterval(ObjectDetectorView view, int overlayInterval) {
        view.setOverlayInterval(overlayInterval);
    }

    @ReactProp(name = "detectionCountBeforeCapture")
    public void setDetectionCountBeforeCapture(ObjectDetectorView view, int count) {
        view.setDetectionCountBeforeCapture(count);
    }

    @ReactProp(name = "overlayColor")
    public void setOverlayColor(ObjectDetectorView view, String colorString) {
        view.setOverlayColor(EngineUtils.HexToColor(colorString));
    }
}
