import { NativeModules, requireNativeComponent, View, UIManager, Platform, PermissionsAndroid, DeviceEventEmitter, Image } from 'react-native';

import React, { Component } from 'react';
import PropTypes from 'prop-types';

import { findNodeHandle } from 'react-native';

const ObjectDetectorView = requireNativeComponent('ObjectDetectorView', ObjectDetector);

import RNFetchBlob from 'rn-fetch-blob';

class ObjectDetector extends Component {
  constructor(props) {
    super(props)
	  this.state = {
      cameraPermissed: Platform.OS === 'ios' ? true : PermissionsAndroid.PERMISSIONS.CAMERA,
      sampleImageURI: ''
    }
  }

  componentDidMount = () => {
    if (Platform.OS === 'android') {
      this.requestCameraPermissions();
      this.detectSubscription = DeviceEventEmitter.addListener('onObjectCaptured', e => {
        this.props.onObjectCaptured && this.props.onObjectCaptured(e.payload);
      });
    } else if (Platform.OS == 'ios') {

    }

    this._updateSampleImage(this.props.sampleImage);
  }

  _onDetectObject = (e) => {
    this.props.onObjectCaptured && this.props.onObjectCaptured(e.nativeEvent.payload);
  }

  _updateSampleImage = (img) => {
    let imgSource = img;
    if (imgSource && typeof imgSource == "number") {
      imgSource = Image.resolveAssetSource(imgSource);
    }

    this.setState({
      sampleImageURI: imgSource.uri
    })
  }

  shouldComponentUpdate(nextProps, nextState) {
    if (this.props.sampleImage != nextProps.sampleImage) {
      this._updateSampleImage(nextProps.sampleImage);
    }

    return true;
  }

  componentWillUnmount = () => {
    if (Platform.OS === 'android') {
      this.detectSubscription && this.detectSubscription.remove();
    }
  }

  async requestCameraPermissions() {
    try {
      let granted = false
      if (this.props.useStorage) {
          granted = await PermissionsAndroid.requestMultiple([
          PermissionsAndroid.PERMISSIONS.CAMERA,
          PermissionsAndroid.PERMISSIONS.WRITE_EXTERNAL_STORAGE,
          PermissionsAndroid.PERMISSIONS.READ_EXTERNAL_STORAGE
          ])
        }
      else {
          granted = await PermissionsAndroid.request(PermissionsAndroid.PERMISSIONS.CAMERA)	
      }
      
      if (granted === PermissionsAndroid.RESULTS.GRANTED) {
        const retJson = await NativeModules.ObjectDetectorModule.initCamera(findNodeHandle(this))
        if (retJson.cameraInitialized) {
              this.setState({cameraPermissed:true})
        }
      } else {
        console.log('Camera permission denied')
      }
    } catch (err) {
      console.warn(err)
    }
  }

  setOverlay(overlayMat) {
    if (Platform.OS === 'android') {
        UIManager.dispatchViewManagerCommand(
          findNodeHandle(this),
          UIManager.getViewManagerConfig('ObjectDetectorView').Commands.setOverlay,
          [overlayMat],
        )
      }
    else {
      const options = { 'overlayMat' : overlayMat }
      NativeModules.ObjectDetectorView.setOverlay(options, findNodeHandle(this))
    }
  }

  async takePicture(filename) {
    var dirs = RNFetchBlob.fs.dirs;

    const outputFilename = dirs.PictureDir + '/' + filename
    const pictureOptions = { 'filename' : outputFilename }

    if (Platform.OS === 'android') {
      return await NativeModules.ObjectDetectorModule.takePicture(pictureOptions, findNodeHandle(this))
    }
    else {
      return await NativeModules.ObjectDetectorView.takePicture(pictureOptions, findNodeHandle(this))
    }
  }

  render() {
    if (!this.state.cameraPermissed) return (<View/>)

    if (Platform.OS === 'android') {
      return (
      <ObjectDetectorView {...this.props} 
        sampleImageURI={ this.state.sampleImageURI } />
      );
    } else if (Platform.OS === 'ios') {
      return (
      <ObjectDetectorView {...this.props} 
        sampleImageURI={ this.state.sampleImageURI } 
        onObjectCaptured={this._onDetectObject}/>
      );
    }

    return (<View/>);
  }
}

ObjectDetector.propTypes = {
  ...View.propTypes,
  detectionCountBeforeCapture: PropTypes.number,
  overlayColor: PropTypes.string
};

export default ObjectDetector;
