#import "ObjectDetector.h"
#import "ObjectDetectorView.h"
#import "EngineUtils.h"
#import <React/RCTBridge.h>
#import <React/RCTUIManager.h>
#import <React/RCTEventDispatcher.h>
#import <React/RCTLog.h>
#import <React/RCTUtils.h>
#import <React/UIView+React.h>

@implementation ObjectDetectorView

RCT_EXPORT_MODULE(ObjectDetectorView)

RCT_EXPORT_VIEW_PROPERTY(onFrameSize, RCTBubblingEventBlock)

RCT_EXPORT_VIEW_PROPERTY(onPayload, RCTBubblingEventBlock)

RCT_EXPORT_VIEW_PROPERTY(onObjectCaptured, RCTBubblingEventBlock)

// And you sent event you want from objectC to react-native
- (UIView *)view {
    return [[ObjectDetector alloc] initWithBridge:self.bridge];
}

RCT_CUSTOM_VIEW_PROPERTY(cvinvoke, NSDictionary*, ObjectDetector) {
    view.mCvInvokeGroup = json;
}

RCT_CUSTOM_VIEW_PROPERTY(facing, NSString*, ObjectDetector) {
    [view changeFacing:json];
}

RCT_CUSTOM_VIEW_PROPERTY(overlay, NSDictionary*, ObjectDetector) {
    view.mOverlay = json;
}

RCT_CUSTOM_VIEW_PROPERTY(overlayInterval, NSNumber*, ObjectDetector) {
    view.mOverlayInterval = json;
}

RCT_CUSTOM_VIEW_PROPERTY(detectionCountBeforeCapture, NSNumber*, ObjectDetector) {
    view.mNDetectionB4Take = [json intValue];
}

RCT_CUSTOM_VIEW_PROPERTY(overlayColor, NSString*, ObjectDetector) {
    [view setOverlayColor:json];
}

RCT_CUSTOM_VIEW_PROPERTY(sampleImageURI, NSString*, ObjectDetector) {
    [view setSampleImage:json];
}

RCT_REMAP_METHOD(setOverlay,
                 options:(NSDictionary *)options
                 reactTag:(nonnull NSNumber *)reactTag)
{
    [self.bridge.uiManager addUIBlock:^(__unused RCTUIManager *uiManager, NSDictionary<NSNumber *, ObjectDetector *> *viewRegistry) {
        ObjectDetector *view = viewRegistry[reactTag];
        if (![view isKindOfClass:[ObjectDetector class]]) {
            RCTLogError(@"Invalid view returned from registry, expecting ObjectDetector, got: %@", view);
        }
        else {
            NSDictionary *overlayMat = [options valueForKey:@"overlayMat"];
            view.mOverlay = overlayMat;
        }
    }];
}

RCT_REMAP_METHOD(takePicture,
                 options:(NSDictionary *)options
                 reactTag:(nonnull NSNumber *)reactTag
                 resolver:(RCTPromiseResolveBlock)resolve
                 rejecter:(RCTPromiseRejectBlock)reject)
{
    [self.bridge.uiManager addUIBlock:^(__unused RCTUIManager *uiManager, NSDictionary<NSNumber *, ObjectDetector *> *viewRegistry) {
        ObjectDetector *view = viewRegistry[reactTag];
        if (![view isKindOfClass:[ObjectDetector class]]) {
            RCTLogError(@"Invalid view returned from registry, expecting ObjectDetector, got: %@", view);
        }
        else {
            TakePicBlock takePicBlock = ^(MatWrapper* MW) {
                [EngineUtils matToImage:MW outPath:[options valueForKey:@"filename"] resolver:resolve rejecter:reject];
            };
            
            [view takePicture:takePicBlock];
        }
    }];
}

@end
